#!/usr/bin/env python3
"""
Power Monitoring for Raspberry Pi
Written by Scott W. Vincent. This was my first Python program and it probably looks like it.
March 2017
http://www.swvincent.com/power-temp-monitor
"""

import RPi.GPIO as GPIO
from datetime import datetime, timedelta
from smtplib import SMTP
from email.mime.text import MIMEText
from configparser import ConfigParser
from retrying import retry
import logging
import signal
import sys
import time
import ds18b20
# import pdb; pdb.set_trace()


def exit_script(signum, frame):
    """Cleanup and exit"""
    GPIO.cleanup()
    logger.info('Program stopped (signal {})'.format(signum))
    sys.exit(0)


def setup_logger():
    """Setup loggers for file and screen"""
    global logger

    # Create logger
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')

    # Log to file
    fh = logging.FileHandler('powmon.log')
    fh.setLevel(logging.INFO)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    # Log to screen
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    logger.info('Program started')


def setup_gpio():
    """Setup GPIO pins"""
    GPIO.setmode(GPIO.BCM)
    GPIO.setup(GREEN_LED, GPIO.OUT)
    GPIO.setup(RED_LED, GPIO.OUT)
    GPIO.setup(PWR_SENSE, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)


def process_power_state():
    """Process power state"""
    global last_power_state, last_power_state_datetime
    power_state = GPIO.input(PWR_SENSE)

    if last_power_state != power_state:
        # Debounce input. Some power supplies may cause a "bounce" to occur
        # as the caps drain, etc. This doesn't really matter except that it
        # creates extra power out/power restored entries in the log file.
        time.sleep(0.02)
        power_state = GPIO.input(PWR_SENSE)
        if last_power_state != power_state:
            # Update status and datetime for last event and reset notification
            last_power_state = power_state
            last_power_state_datetime = datetime.now()

            # Record it
            if last_power_state:
                logger.info('Power restored')
            else:
                logger.info('Power lost')

    # Update LEDs. I do this even if there's no change, otherwise they get missed
    # when program is started if power state is the same as last time.
    GPIO.output(GREEN_LED, power_state)
    GPIO.output(RED_LED, not power_state)


def process_temp():
    """Process temperature reading"""
    global last_temp_state, last_temp_state_datetime

    temp_state_timedelta = datetime.now() - last_temp_state_datetime
    if temp_state_timedelta.seconds > TEMP_TIMER:
        temps = ds18b20.read_temp()
        if temps is not None:
            # I use Fahrenheit because 'Murica. Change 1 to 0 if you prefer Celsius.
            temp = temps[1]
            logger.debug('Temperature is {:.1f} degrees'.format(temp))
            if last_temp_state is True:
                # Temp was previously under control and remains so if we stay under warning temp.
                temp_state = temp < WARNING_TEMP
            else:
                # Temp was previously too high and remains so until we drop below target temp.
                temp_state = temp < TARGET_TEMP

            if last_temp_state != temp_state:
                # Update status for last event and reset notification
                last_temp_state = temp_state

                # Record it
                if last_temp_state:
                    logger.info('Temperature under control ({:.1f} degrees)'.format(temp))
                else:
                    logger.info('Temperature too high ({:.1f} degrees)'.format(temp))
        else:
            # I just write as a debug so log file doesn't get overwhelmed
            logger.debug('Temperature could not be read')

        last_temp_state_datetime = datetime.now()


def process_power_notifications():
    """Send power notification email if conditions met"""
    global power_off_notification_sent
    # Check time since power state last changed
    power_state_timedelta = datetime.now() - last_power_state_datetime
    power_state_seconds = power_state_timedelta.seconds

    if power_off_notification_sent and last_power_state and power_state_seconds >= POWER_ON_TIMER:
        # Power is on and power out notification has previously been sent and time has passed
        try:
            send_email(email_list_power, 'Power Restored at {}'.format(LOCATION), 'Power has been restored at {} as of {:%-m/%-d/%Y %-I:%M %p}.\n{}'.format(LOCATION,
                       last_power_state_datetime, FOOTER_TEXT))
        except Exception as ex:
            # Email failed to send after several retries
            logger.error('Could not send power restored email after several attempts: ' + str(ex))
        else:
            # Reset Power off notification sent, then we're ready for next notification
            power_off_notification_sent = False
            save_settings()
            # logger.info('Email sent: power has been on for {} seconds'.format(power_state_seconds))
            logger.info('Email sent: Power has been restored at {} as of {:%-m/%-d/%Y %-I:%M %p}'.format(LOCATION,
                        last_power_state_datetime))
    elif not power_off_notification_sent and not last_power_state and power_state_seconds >= POWER_OFF_TIMER:
        # Power is off for set amount of time and notification not sent yet, send it
        try:
            send_email(email_list_power, 'Power Outage at {}'.format(LOCATION), 'Power has been out at {} since {:%-m/%-d/%Y %-I:%M %p}.\n{}'.format(LOCATION,
                       last_power_state_datetime, FOOTER_TEXT))
        except Exception as ex:
            # Email failed to send after several retries
            logger.error('Could not send power outage email after several attempts: ' + str(ex))
        else:
            # Indicate that notification was sent
            power_off_notification_sent = True
            save_settings()
            # logger.info('Email sent: power has been out for {} seconds'.format(power_state_seconds))
            logger.info('Email sent: Power has been out at {} as of {:%-m/%-d/%Y %-I:%M %p}'.format(LOCATION,
                        last_power_state_datetime))


def process_temp_notifications():
    """Send temperature too high notification email if conditions met"""
    global temp_too_high_notification_sent

    if temp_too_high_notification_sent and last_temp_state:
        # Temp is under control and temp too high notification has previously been sent
        try:
            send_email(email_list_temp, 'Temp under control at {}'.format(LOCATION), 'Temp is under control (<{}) at {} since {:%-m/%-d/%Y %-I:%M %p}.\n{}'.format(TARGET_TEMP,
                       LOCATION, last_temp_state_datetime, FOOTER_TEXT))
        except Exception as ex:
            # Email failed to send after several retries
            logger.error('Could not send temp under control email after several attempts: ' + str(ex))
        else:
            # Reset temp too high notification sent, then we're ready for next notification
            temp_too_high_notification_sent = False
            save_settings()
            logger.info('Email sent: temp under control')
    elif not temp_too_high_notification_sent and not last_temp_state:
        # Temp is too high and notification has not been sent yet, send it
        try:
            send_email(email_list_temp, 'Temp too high at {}'.format(LOCATION), 'Temp is too high (>{}) at {} since {:%-m/%-d/%Y %-I:%M %p}.\n{}'.format(WARNING_TEMP,
                       LOCATION, last_temp_state_datetime, FOOTER_TEXT))
        except Exception as ex:
            # Email failed to send after several retries
            logger.error('Could not send temp too high email after several attempts: ' + str(ex))
        else:
            # Reset temp too high notification sent, then we're ready for next notification
            temp_too_high_notification_sent = True
            save_settings()
            logger.info('Email sent: temp too high')


@retry(wait_fixed=30000, stop_max_attempt_number=4)
def send_email(email_list, subjectText, msgText):
    """
    This will retry 4 times, every 30 seconds then fail. Main loop can
    then run and process will begin again, so change in power status,
    temperature, etc. doesn't get ignored indefinitely.
    """
    logger.debug('Attempting to send email "{}"'.format(subjectText))
    msg = MIMEText(msgText)
    msg['Subject'] = subjectText
    msg['From'] = FROM_EMAIL_ADDRESS
    msg['To'] = FROM_EMAIL_ADDRESS
    msg['Bcc'] = email_list

    smtp = SMTP(MAIL_SERVER)
    smtp.send_message(msg)
    smtp.quit()


def load_settings():
    """Load settings, or create new settings file if it's missing"""
    global last_power_state, power_off_notification_sent, last_temp_state, temp_too_high_notification_sent

    try:
        config = ConfigParser()
        config.read('powmon.ini')
        last_power_state = config.getboolean('main', 'last_power_state')
        power_off_notification_sent = config.getboolean('main', 'power_off_notification_sent')
        last_temp_state = config.getboolean('main', 'last_temp_state')
        temp_too_high_notification_sent = config.getboolean('main', 'temp_too_high_notification_sent')
        logger.debug('Settings loaded')
    except Exception:
        # If exception occurs, assume file is damaged/missing, rebuild. Assume that power
        # was out and system is coming back online, I suppose file could possibly be
        # corrupted if it was writing file and lost power at that point.
        last_power_state = True
        power_off_notification_sent = True
        last_temp_state = True
        temp_too_high_notification_sent = False

        config = ConfigParser()
        config.add_section('main')
        config.set('main', 'last_power_state', str(last_power_state))
        config.set('main', 'power_off_notification_sent', str(power_off_notification_sent))
        config.set('main', 'last_temp_state', str(last_temp_state))
        config.set('main', 'temp_too_high_notification_sent', str(temp_too_high_notification_sent))

        with open('powmon.ini', 'w') as f:
            config.write(f)

        logger.warning('Settings file rebuilt because it was missing/corrupt')


def save_settings():
    """Save settings"""
    config = ConfigParser()
    config.read('powmon.ini')
    config.set('main', 'last_power_state', str(last_power_state))
    config.set('main', 'power_off_notification_sent', str(power_off_notification_sent))
    config.set('main', 'last_temp_state', str(last_temp_state))
    config.set('main', 'temp_too_high_notification_sent', str(temp_too_high_notification_sent))

    with open('powmon.ini', 'w') as f:
        config.write(f)


def setup_email_lists():
    """Load email lists from files"""
    global email_list_power, email_list_temp

    email_list_power = load_lines_from_file('email-list-power.txt')
    if email_list_power is not None:
        logger.debug('Power outage emails loaded: {}'.format(email_list_power))
    else:
        logger.critical('Power outage emails not loaded, program will exit')
        exit_script(None, None)

    email_list_temp = load_lines_from_file('email-list-temperature.txt')
    if email_list_temp is not None:
        logger.debug('Temperature emails loaded: {}'.format(email_list_temp))
    else:
        logger.critical('Temperature emails could not be loaded, program will exit')
        exit_script(None, None)


def load_lines_from_file(loadfile):
    """Load lines from text file"""
    try:
        with open(loadfile) as f:
            lines = f.read().splitlines()
            lines = list(line for line in lines if line)  # Filter out blank lines
            email_list = ', '.join(lines)
            return email_list
    except Exception as ex:
        logger.error('Exception occured while loading {}. Details: {}'.format(loadfile, str(ex)))


# Global Constants
GREEN_LED = 24
RED_LED = 23
PWR_SENSE = 25
POWER_OFF_TIMER = 600
POWER_ON_TIMER = 900
TEMP_TIMER = 60
TARGET_TEMP = 87
WARNING_TEMP = 92
MAIL_SERVER = 'mail.yourdomain.com'
FROM_EMAIL_ADDRESS = 'noreply@yourdomain.com'
LOCATION = 'YOUR SITE'
FOOTER_TEXT = '*This is an automated message*'

# Global Variables
start_date_time = datetime.now()
last_power_state_datetime = start_date_time
last_temp_state_datetime = start_date_time - timedelta(seconds=TEMP_TIMER)

# Setup Signals
signal.signal(signal.SIGTERM, exit_script)
signal.signal(signal.SIGINT, exit_script)

# Setup functions
setup_logger()
load_settings()
setup_gpio()
setup_email_lists()

while True:
    # Main loop, continues until script is ended by a signal
    process_power_state()
    process_power_notifications()
    process_temp()
    process_temp_notifications()
